use std::{ops::Deref, sync::Arc};

use anyhow::Context;
use sqlx::{Pool, Postgres};

use crate::services::{consumers::Consumers, handler::Handler, staging::Staging};

#[derive(Clone)]
pub struct SharedState(Arc<State>);

impl SharedState {
    pub async fn new() -> anyhow::Result<Self> {
        Ok(Self(Arc::new(State::new().await?)))
    }
}

impl Deref for SharedState {
    type Target = Arc<State>;

    fn deref(&self) -> &Self::Target {
        &self.0
    }
}

pub struct State {
    pub _db: Pool<Postgres>,
    pub staging: Staging,
    pub consumers: Consumers,
    pub handler: Handler,
}

impl State {
    pub async fn new() -> anyhow::Result<Self> {
        let db = sqlx::PgPool::connect(
            &std::env::var("DATABASE_URL").context("DATABASE_URL is not set")?,
        )
        .await?;

        sqlx::migrate!("migrations/crdb")
            .set_locking(false)
            .run(&db)
            .await?;

        let _ = sqlx::query("SELECT 1;").fetch_one(&db).await?;

        let staging = Staging::default();
        let handler = Handler::new(staging.clone());

        Ok(Self {
            _db: db,
            consumers: Consumers::new(),
            staging,
            handler,
        })
    }
}
